import subprocess as sp
import os
from qtpy import QtCore, QtWidgets
import multiprocessing as mp
import time


class Result:
    def __init__(
        self, url, filename, title, text, highlighted_text, search_text, snippets, found_text, type
    ):
        self.url = os.path.abspath(url) if len(url) > 0 else ""
        self.filename = filename
        self.title = title
        self.text = text
        self.highlighted_text = highlighted_text
        self.search_text = search_text
        self.snippets = snippets
        self.found_text = found_text
        self.type = type


class SearchEngineError(ValueError):
    pass


class _ThreadEngine(QtCore.QThread):
    def __init__(self, engine):
        super().__init__()
        self.engine = engine
        self.search_text = ""

    def run(self):
        if self.search_text != "":
            self.engine.search(self.search_text)
        self.finished.emit()


class ThreadedSearch(QtCore.QObject):
    search_finished = QtCore.Signal(list)

    def __init__(self, engines):
        super().__init__()
        self.engines = engines
        self.threads = []
        for e in self.engines:
            e.search_finished.connect(self.engine_finished)
            e.log.connect(self.search_proc)
            self.threads.append(_ThreadEngine(e))

        self.finished_threads = 0
        self.first_search = True
        self.results = []
        self.log = []
        self.start_time = 0

    def search_proc(self, log):
        self.log.append(log)

    def engine_finished(self, results):
        self.results.extend(results)
        self.finished_threads += 1
        if self.finished_threads == len(self.engines):
            self.search_finished.emit(self.results)
            # print(time.perf_counter() - self.start_time)

    def search(self, text):
        self.start_time = time.perf_counter()
        self.results = []
        self.finished_threads = 0
        self.log = []
        for t in self.threads:
            t.search_text = text
            t.start()

    def stop(self):
        for e, t in zip(self.engines, self.threads):
            t.quit()
            e.stop()
            t.wait()
            t.deleteLater()


class Engine(QtCore.QObject):
    search_finished = QtCore.Signal(list)
    log = QtCore.Signal(str)
    available_engines = {}
    args = ["path", "max_found"]
    engines = {}

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.available_engines[cls.__name__.lower()] = cls
        cls.name = cls.__name__[: -len("search")]

    @classmethod
    def create(cls, name, **kwargs):
        if isinstance(name, (list, tuple)):
            # return ChainedEngine([cls.create(i, **kwargs) for i in name])
            return ThreadedSearch([cls.create(i, **kwargs) for i in name])
        if name + "search" in cls.engines:
            e = cls.engines[name + "search"]
            if "path" in kwargs:
                e.search_path = kwargs["path"]
            if "max_found" in kwargs:
                e.max_found = kwargs["max_found"]
            return e
        if name not in cls.available_engines:
            name = name + "search"
            if name not in cls.available_engines:
                raise SearchEngineError(f"Search engine '{name}' does not exist")
        engine = cls.available_engines[name]
        for k in list(kwargs.keys()):  # make sure the class gets only what it wants
            if k not in engine.args:
                del kwargs[k]
        e = engine(**kwargs)
        cls.engines[name] = e
        return e

    def __init__(self, path, max_found, **kwargs):
        super().__init__()
        self.search_path = path
        self.max_found = max_found
        self.timeout = None
        self._proc = None

    def run_external(self, cmd, **kwargs):
        self._proc = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.DEVNULL, **kwargs)
        try:
            out, _ = self._proc.communicate(timeout=self.timeout)
        except sp.TimeoutExpired as e:
            self._proc.kill()
            out = e.output
        self._proc = None
        return out

    def stop(self):
        if self._proc is not None:
            self._proc.terminate()
            try:
                self._proc.wait(3)
            except sp.TimeoutExpired:
                self._proc.kill()

    def search(self, text):
        self.log.emit(f"Started search with engine {self.__class__.__name__}")
        res = self.do_search(text)
        self.log.emit(f"Finished search with engine {self.__class__.__name__}")
        self.search_finished.emit(res)

    def do_search(self, text):
        raise NotImplementedError(
            f"This search engine does not implement search {self.__class__.__name__}"
        )


# class ChainedEngine(Engine):
#     def __init__(self, engines):
#         self.engines = engines

#     def search(self, text):
#         results = []
#         for e in self.engines:
#             results.append(e.search(text))
#         return results


class SplittableHighlighter:
    def startMatch(self, index):
        return "##|##"

    def endMatch(self):
        return "||#||"


class Highlighter:
    def startMatch(self, index):
        return '<font color="red">'

    def endMatch(self):
        return "</font>"


class RecollSearch(Engine):
    args = ["max_found"]

    def __init__(self, max_found):
        super().__init__("", max_found)
        from recoll import recoll

        self.recoll = recoll.connect()
        self.query = self.recoll.query()
        self.highlighter = Highlighter()
        self.filter = " issub:0"

    def do_search(self, text):
        # num = self.query.execute(text + " mime:application/pdf issub:0")
        num = self.query.execute(text + self.filter)
        results = []
        i = 0
        while i < min(num, self.max_found):
            doc = self.query.fetchone()
            url = doc.get("url")
            if True:  # url.lower().endswith("pdf"):
                i += 1
                try:
                    found_text = (
                        self.query.makedocabstract(doc, SplittableHighlighter())
                        .split("##|##", 1)[1]
                        .replace("||#|| ##|##", " ")
                        .split("||#||", 1)[0]
                    )
                except IndexError:
                    found_text = text

                if doc.get("mtype").endswith("pdf"):
                    mime = "pdf"
                elif "text" in sp.check_output(["file", url.split("//", 1)[1]]).decode():
                    mime = "text"
                elif doc.get("mtype").startswith("image"):
                    mime = "image"
                else:
                    mime = "untreated"

                results.append(
                    Result(
                        url.split("//", 1)[1],
                        doc.get("filename"),
                        doc.get("title"),
                        self.query.makedocabstract(doc),
                        self.query.makedocabstract(doc, self.highlighter),
                        text,
                        self.query.getsnippets(doc, methods=SplittableHighlighter()),
                        found_text,
                        # "pdf" if doc.get("mtype").endswith("pdf") else doc.get("mtype"),
                        mime,
                    )
                )
            else:
                print("skipped", url)
        return results


class RecollPDFSearch(RecollSearch):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter += " mime:application/pdf"


class PDFGrepSearch(Engine):
    args = Engine.args + ["timeout"]

    def __init__(self, timeout=None, **kwargs):
        super().__init__(**kwargs)
        self.timeout = timeout

    def do_search(self, text):
        output = self.run_external(
            ["pdfgrep", text, "-r", "-n", "--cache", "-i", "-m", "10"],
            cwd=self.search_path,  # , "-C", "2"],
        ).decode()
        matches = output.split("\n")
        results = []
        # current = []
        # cfn = "NOFILENAME"
        # found = []
        # search_with_context = False
        # if search_with_context:
        # for match in matches:
        # if match.strip() == "":
        # found.append(current)
        # break
        # if match.startswith(cfn):
        # if len(match) > len(cfn) + 2 + len(current[1]):
        # current[2] += match[len(cfn) + 2 + len(current[1]) :]
        # else:
        # current[2] += "\n"
        # if "pdf:" in match:
        # current.append(match[len(cfn) + 2 + len(current[1]) :])
        # else:
        # if len(found) == self.max_found:
        # break
        # if cfn != "NOFILENAME" or match == "--":
        # found.append(current)
        # if match == "--":
        # cfn = "NOFILENAME"
        # current = []
        # continue
        # if "pdf:" in match:
        # current = match.split(":", 2)
        # current.append(current[2])
        # else:
        # fname, rest = match.split("pdf-", 1)
        # fname += "pdf"
        # lineno, ftext = rest.split("-", 1)
        # current = [fname, lineno, ftext]
        # cfn = current[0]
        # else:
        found = list(map(lambda x: x.split(":", 2), matches[:-1]))

        for f in found[: self.max_found]:
            path, page, ftext = f
            try:
                i = ftext.lower().index(text.lower())
                ftext = ftext[:i].split("   ")[-1] + ftext[i:].split("   ")[0]
                i = ftext.lower().index(text.lower())
                term = ftext[max(0, i - 5) : i + len(text) + 5].strip()
            except ValueError:
                term = text

            results.append(
                Result(
                    os.path.join(self.search_path, path),
                    os.path.basename(path),
                    "",
                    ftext,
                    f"[p {page}] {ftext}",
                    text,
                    [(int(page), None, None)],
                    term,
                    "pdf",
                )
            )

        return results


class AGSearch(Engine):
    args = ["path", "max_found", "skip_filenamebased", "timeout"]

    def __init__(self, path, max_found, skip_filenamebased=False, timeout=None):
        super().__init__(path, max_found)
        self.timeout = timeout
        self.skip_filenamebased = skip_filenamebased
        if sp.call(["which", "ag"], stdout=sp.DEVNULL, stderr=sp.DEVNULL) == 0:
            self.prog = ["ag", "-m", "10"]
            print("Using silver searcher")
        else:
            self.prog = ["grep", "-n", "-r", "-m", "10"]
            print("Using grep")

    def do_search(self, text):
        output = self.run_external(self.prog + [text], cwd=self.search_path)
        matches = output.strip().split(b"\n")

        def split(x):
            # Note --null in ag does not work when piping or working with subprocess
            if x == b"":
                return [x]
            f = x.split(b":", 1)
            while f[1][0] not in b"0123456789":
                s = f[1].split(b":", 1)
                try:
                    f = [f[0] + s[0], s[1]]
                except:
                    raise ValueError(str([x, f, s]))
            b, t = f[1].split(b":", 1)

            return [f[0], b, t]

        found = list(map(split, matches))
        results = {}
        i = 0
        k = 0
        while i < self.max_found and k < len(found):
            f = found[k]
            k += 1
            if f == [b""]:
                continue
            path, line, ftext = f
            path = path.decode()
            line = line.decode()
            ftext = ftext.decode()
            try:
                term = text
                if path in results:
                    results[path].highlighted_text += f"[l {line}] {ftext}"
                    results[path].text += ftext
                else:
                    results[path] = Result(
                        os.path.join(self.search_path, path),
                        os.path.basename(path),
                        "",
                        ftext,
                        f"[l {line}] {ftext}",
                        text,
                        [(int(line), None, None)],
                        term,
                        "text",
                    )
                    i += 1
            except UnicodeDecodeError:
                pass

        # if not self.skip_filenamebased:
        #     out = sp.run(
        #         ["find", self.search_path, "-iname", f"*{text}*", "-type", "f"],
        #         check=False,
        #         capture_output=True,
        #     )
        #     out = out.stdout.decode().strip()

        #     def check_mime(p):
        #         t = sp.check_output(["file", p]).decode()
        #         if "text" in t:
        #             return "text"
        #         elif "image" in t:
        #             return "image"
        #         elif "pdf" in t:
        #             return "pdf"
        #         else:
        #             return "untreated"

        #     for line in out.strip().split("\n")[: self.max_found]:
        #         if line == "":
        #             continue
        #         if line not in results:
        #             results[line] = Result(
        #                 line,
        #                 os.path.basename(line),
        #                 "",
        #                 "",
        #                 "",
        #                 text,
        #                 [(0, 0, 0)],
        #                 "",
        #                 check_mime(line),
        #             )
        return list(results.values())


def check_mime(p):
    file_output = sp.check_output(["file", p]).decode()
    if "text" in file_output:
        return "text"
    elif "pdf" in file_output:
        return "pdf"
    elif "image" in file_output:
        return "image"
    else:
        return "untreated"


class FileNameSearch(Engine):
    args = Engine.args + ["timeout"]

    def __init__(self, timeout=None, **kwargs):
        super().__init__(**kwargs)
        self.timeout = timeout

    def do_search(self, text):
        results = []
        out = self.run_external(["find", self.search_path, "-iname", f"*{text}*", "-type", "f"])

        out = out.decode()
        for line in out.strip().split("\n")[: self.max_found]:
            if line == "":
                continue
            results.append(
                Result(
                    line,
                    os.path.basename(line),
                    "",
                    "",
                    "",
                    text,
                    [(0, 0, 0)],
                    "",
                    check_mime(line),
                )
            )
        return results


# class AG_PDFGrepSearch(Engine):
#     def __init__(self, path, max_found, engine):
#         self.pdfgrep = PDFGrepSearch(path, max_found // 2)
#         self.ag = AGSearch(path, max_found // 2, skip_filenamebased=True)
#         self.engine = engine

#     def search(self, text):
#         if self.engine == "ag+pdfgrep":
#             result = self.ag.search(text)
#             result += self.pdfgrep.search(text)
#         else:
#             result = self.pdfgrep.search(text)
#             result += self.ag.search(text)
#         return result
