#!/bin/env python3
import enum
import os
import sys
import time
import logging

from .log import LogEnabled


# The logic is following:
#   We start normally and check if we are in Qt6
#   If not, we try to use PySide6 and restart the process
#   If yes, we continue
#   If we restarted and we cannot import qtpy (because PySide6 is not available)
#   We try to restart with PyQt6
#   If both does not work we stop

try:
    import qtpy
except:
    if os.environ.get("pyside_tested", False):
        if os.environ.get("pyqt_tested", False):
            logging.error(
                "Cannot find suitable version of QT bindings, install either PySide6 or PyQt6"
            )
            exit()
        os.environ["QT_API"] = "PyQt6"
        os.environ["pyqt_tested"] = "1"
        os.execv(sys.executable, ["python3"] + sys.argv)

if not qtpy.QT6:
    os.environ["QT_API"] = "PySide6"
    os.environ["pyside_tested"] = "1"
    os.execv(sys.executable, ["python3"] + sys.argv)

from qtpy import QtWidgets, QtCore, QtGui
from qtpy import QtPdf, QtPdfWidgets
import typing
import subprocess as sp
from configparser import ConfigParser
from .engines import Engine, Result, check_mime
from .tagging import Tagger, TagWidget


class Config:
    _config = ConfigParser()
    defaults = {
        "Search": {"timeout": 15, "engines": "pdfgrep", "path": os.path.expanduser("~/Documents")},
        "Results": {
            "limit_num": 100,
            "icon_width": 30,
            "tag_database_path": os.path.join(
                os.path.expanduser("~"), ".config", "pdfsearch_tags.db"
            ),
            "comments_as_markdown": False,
            "show_result_status": False,
            "revert_to_search_results": False,
        },
        "Optics": {
            "page_button_order": "okular",
            "show_list_separator": True,
            "list_separator_px": 1,
            "dirtree_shown": False,
            "dirtree_width_percent": 30,
            "viewer_width_percent": 60,
            "text_highlight_bg_color": "red",
            "text_highlight_fg_color": "white",
            "tags_shown": True,
            "tags_width_percent": 15,
        },
        "Optics.FontSizes": {"filename_size": 12, "context_size": 10},
    }

    @classmethod
    def read(cls, path):
        cls._config.update(cls.defaults)
        if os.path.isfile(path):
            cls._config.read(path)
        cls.file = path

    @classmethod
    def _bool(cls, value):
        if isinstance(value, bool):
            return value
        if value.lower() in ["off", "false"]:
            return False
        return bool(value)

    @classmethod
    def get(cls, section, option, type=str):
        # default = cls.defaults.get(section, {}).get(option, None)
        default = None
        if type == bool:
            type = cls._bool
        if section in cls._config:
            val = cls._config[section].get(option, default)
            if isinstance(val, str) and "," in val:
                return list(map(lambda x: type(x.strip()), val.split(",")))
            return type(val)
        else:
            return default
            # try:
            # return type(default)
            # except:
            # return None

    @classmethod
    def set(cls, section, option, value):
        cls._config.setdefault(section, {})
        cls._config[section][option] = str(value)


class WrapLabel(QtWidgets.QLabel):
    def __init__(self, text, *args, **kwargs):
        text = "\u200B".join(text)
        super().__init__(text, *args, **kwargs)
        self.setWordWrap(True)


class ClickLabel(WrapLabel):
    clicked = QtCore.Signal()

    def mousePressEvent(self, ev):
        self.clicked.emit()


class RectPDF(QtPdfWidgets.QPdfView):
    def __init__(self, parent):
        super().__init__(parent)
        self.sindex = 0
        palette = self.palette()
        dark = palette.color(QtGui.QPalette.Dark)
        base = palette.color(QtGui.QPalette.Base)
        if dark.lightness() < base.lightness():
            palette.setBrush(QtGui.QPalette.Dark, dark)  # QtGui.QColor("lime"))
        else:
            palette.setBrush(QtGui.QPalette.Dark, base)  # QtGui.QColor("lime"))
        self.setPalette(palette)

    def scroll_perc(self):
        vs = self.verticalScrollBar()
        mi = vs.minimum()
        ma = vs.maximum()
        val = vs.value()
        if mi == ma:
            return 1
        return val / (ma - mi)

    def scroll_to_perc(self, perc):
        vs = self.verticalScrollBar()
        mi = vs.minimum()
        ma = vs.maximum()
        val = perc * (ma - mi)
        vs.setValue(int(val))

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        p = self.scroll_perc()
        r = super().resizeEvent(event)
        self.scroll_to_perc(p)
        return r

    def setZoomFactor(self, factor: float) -> None:
        p = self.scroll_perc()
        super().setZoomFactor(factor)
        self.scroll_to_perc(p)

    def setCurrentSearchResultIndex(self, index):
        if index >= 0:
            self.sindex = index
            link = self.searchModel().resultAtIndex(index)
            for l in link.rectangles():
                l.adjust(-2, -2, 2, 200)
            self.pageNavigator().jump(link)
        super().setCurrentSearchResultIndex(index)

    def wheelEvent(self, event):
        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                self.setZoomFactor(min(2, self.zoomFactor() + 0.05))
            else:
                self.setZoomFactor(max(0.15, self.zoomFactor() - 0.05))
        else:
            super().wheelEvent(event)


class FlatHoverButton(QtWidgets.QPushButton):
    def __init__(self, iconname, text="", alt=""):
        super().__init__(text)
        ico = QtGui.QIcon.fromTheme(iconname)
        self.setIcon(ico)
        if ico.isNull():
            self.setStyleSheet("font-size: 11pt")
            self.setText(alt)
        if not text:
            self.setFixedWidth(30)
        self.setFlat(True)

    def event(self, e: QtCore.QEvent) -> bool:
        if e.type() == QtCore.QEvent.HoverEnter and self.isEnabled():
            self.setFlat(False)
        elif e.type() == QtCore.QEvent.HoverLeave:
            self.setFlat(True)
        return super().event(e)

    def setDisabled(self, arg__1: bool) -> None:
        if arg__1:
            self.setFlat(True)
        return super().setDisabled(arg__1)


class ZoomWidget(QtWidgets.QWidget):
    zoomChanged = QtCore.Signal(float)

    def __init__(self, zoom):
        super().__init__()
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.less = FlatHoverButton("zoom-out", alt="-")
        self.more = FlatHoverButton("zoom-in", alt="+")

        self.value = QtWidgets.QSpinBox()
        self.value.setMaximumWidth(50)
        self.value.setMaximum(200)
        self.value.setMinimum(15)
        self.layout.addWidget(QtWidgets.QLabel(" Zoom "))
        self.layout.addWidget(self.less)
        self.layout.addWidget(self.value)
        self.layout.addWidget(self.more)
        self.zoom = zoom
        self.value.setValue(int(self.zoom * 100))

        self.less.clicked.connect(lambda: self.set_zoom("-"))
        self.more.clicked.connect(lambda: self.set_zoom("+"))
        self.value.valueChanged.connect(lambda val: self.set_zoom(value=val))
        self.setMaximumHeight(40)
        self.layout.setSpacing(0)
        self.layout.invalidate()

    def set_zoom(self, mode=None, value=None):
        if mode == "+":
            self.zoom = min(2, self.zoom + 0.05)
        elif mode == "-":
            self.zoom = max(0.15, self.zoom - 0.05)
        else:
            self.zoom = value / 100
        self.zoomChanged.emit(self.zoom)
        self.value.blockSignals(True)
        self.value.setValue(round(self.zoom * 100))
        self.value.blockSignals(False)
        self.less.setDisabled(self.zoom == 0.15)
        self.more.setDisabled(self.zoom == 2)


class PageWidget(QtWidgets.QWidget):
    def __init__(self, nav: QtPdf.QPdfPageNavigator, doc: QtPdf.QPdfDocument):
        super().__init__()
        self.nav = nav
        self.doc = doc
        self.page_count = doc.pageCount()
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.prev = FlatHoverButton("go-down", alt="-")
        self.next = FlatHoverButton("go-up", alt="+")
        self.val = QtWidgets.QSpinBox()
        self.val.setFixedWidth(50)
        self.prev.clicked.connect(lambda: self.update_page(mode="-"))
        self.next.clicked.connect(lambda: self.update_page(mode="+"))
        self.val.valueChanged.connect(lambda v: self.update_page(value=v))
        self.val.setMinimum(1)
        self.nav.currentPageChanged.connect(self.page_changed)
        self.doc.pageCountChanged.connect(self.set_page_count)
        self.add_info_label = QtWidgets.QLabel("")
        self.layout.addWidget(QtWidgets.QLabel(" Page "))
        if Config.get("Optics", "page_button_order") == "natural":
            self.layout.addWidget(self.prev)
            self.layout.addWidget(self.val)
            self.layout.addWidget(self.add_info_label)
            self.layout.addWidget(self.next)
        elif Config.get("Optics", "page_button_order") == "okular":
            self.layout.addWidget(self.next)
            self.layout.addWidget(self.val)
            self.layout.addWidget(self.add_info_label)
            self.layout.addWidget(self.prev)
        else:
            raise ValueError("Wrong configuration for page button order")
        self.setMaximumHeight(40)
        self.layout.setSpacing(0)
        self.layout.invalidate()

    def set_page_count(self, val):
        self.page_count = val
        self.val.setMaximum(val)
        self.prev.setDisabled(self.nav.currentPage() == 0)
        self.next.setDisabled(self.nav.currentPage() == self.page_count - 1)
        # self.val.setSuffix(f" / {self.page_count}")
        self.set_add_info()

    def set_add_info(self):
        cp = self.nav.currentPage()
        text = ""
        if self.doc.pageLabel(cp) != str(cp + 1):
            text = f"({self.doc.pageLabel(cp)})"
        text += f" / {self.page_count}"
        self.add_info_label.setText(text)

    def page_changed(self, val):
        self.val.blockSignals(True)
        self.val.setValue(val + 1)
        self.set_add_info()
        self.prev.setDisabled(self.nav.currentPage() == 0)
        self.next.setDisabled(self.nav.currentPage() == self.page_count - 1)
        self.val.blockSignals(False)

    def update_page(self, mode=None, value=None):
        if mode == "+":
            self.nav.jump(
                min(self.page_count, self.nav.currentPage() + 1),
                QtCore.QPointF(0, 0),
                self.nav.currentZoom(),
            )
        elif mode == "-":
            self.nav.jump(
                max(0, self.nav.currentPage() - 1),
                self.nav.currentLocation(),
                self.nav.currentZoom(),
            )
        else:
            self.nav.jump(value - 1, QtCore.QPointF(0, 0), self.nav.currentZoom())

        self.nav.parent().setCurrentSearchResultIndex(-1)

        self.prev.setDisabled(self.nav.currentPage() == 0)
        self.next.setDisabled(self.nav.currentPage() == self.page_count - 1)


class PDF(LogEnabled, QtWidgets.QWidget):
    def __init__(self, initial_zoom):
        super().__init__()
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.list = QtWidgets.QListView()
        self.list.setMaximumWidth(150)
        self.list.hide()

        self.view = RectPDF(self)
        self.view.installEventFilter(self)

        self.split = QtWidgets.QSplitter()
        self.split.addWidget(self.list)
        self.split.addWidget(self.view)
        upperLayout = QtWidgets.QHBoxLayout()
        self.layout.addLayout(upperLayout)
        upperLayout.addWidget(self.split)

        self.zoomWidget = ZoomWidget(self.view.pageNavigator().currentZoom())
        self.zoomWidget.zoomChanged.connect(self.view.setZoomFactor)

        self.document = QtPdf.QPdfDocument(self.view)
        self.pageWidget = PageWidget(self.view.pageNavigator(), self.document)

        lowerLayout = QtWidgets.QHBoxLayout()

        self.search_prev_button = FlatHoverButton("go-previous", alt="<")
        self.search_prev_button.clicked.connect(lambda: self.search_next(dir=-1))
        self.search_prev_button.setDisabled(True)
        self.search_next_button = FlatHoverButton("go-next", alt=">")
        self.search_next_button.clicked.connect(self.search_next)
        self.search_next_button.setDisabled(True)
        self.search_label = QtWidgets.QLabel("0")

        lowerLayout.addWidget(QtWidgets.QLabel("Search "))
        lowerLayout.addWidget(self.search_prev_button)
        lowerLayout.addWidget(self.search_label)
        lowerLayout.addWidget(self.search_next_button)

        lowerLayout.addStretch(0)
        lowerLayout.addWidget(self.zoomWidget)
        # lowerLayout.addStretch(0)
        lowerLayout.addWidget(self.pageWidget)
        lowerLayout.addStretch(0)
        self.open_ext_button = FlatHoverButton("document-open", alt="Open")
        self.open_ext_button.clicked.connect(self.open_external)
        self.open_ext_button.setDisabled(True)
        self.open_ext_button.setToolTip("Open Externally")
        lowerLayout.addWidget(self.open_ext_button)

        self.layout.addLayout(lowerLayout)

        self.view.setPageMode(QtPdfWidgets.QPdfView.PageMode.MultiPage)
        self.view.setDocument(self.document)
        self.model = QtPdf.QPdfSearchModel(self.view)
        try:
            self.model.countChanged.connect(self.update_search_label)
            self.needstimer = False
        except:
            # Qt 6.8 will introduce the signal
            self.needstimer = True
        self.view.setSearchModel(self.model)
        self.model.setDocument(self.document)
        self.nav = self.view.pageNavigator()
        self.list.setModel(self.model)
        self.list.selectionModel().currentChanged.connect(
            lambda c, p: self.view.setCurrentSearchResultIndex(c.row())
        )
        self.current_file = None
        self.zoomWidget.set_zoom(value=initial_zoom)
        self.view.zoomFactorChanged.connect(self.zoomchange)
        # only necessary on qt < 6.8
        if self.needstimer:
            self.numsearstart = time.perf_counter()
            self.numsearchTimer = QtCore.QTimer()
            self.numsearchTimer.setInterval(200)
            self.numsearchTimer.timeout.connect(self.update_search_label)

        self.lowerLayout = lowerLayout

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        return super().resizeEvent(event)

    def open_external(self):
        sp.call(["xdg-open", self.current_file])

    def zoomchange(self, fac):
        self.zoomWidget.blockSignals(True)
        self.zoomWidget.set_zoom(value=fac * 100)
        self.zoomWidget.blockSignals(False)

    def eventFilter(self, watched, event):
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Left:
                self.pageWidget.update_page(mode="-")
            elif event.key() == QtCore.Qt.Key_Right:
                self.pageWidget.update_page(mode="+")
        return False

    def open(self, filename):
        self.log.debug(f"PDF: Opening File {filename}")
        self.model.setSearchString("")
        if filename != self.current_file or self.document.status() in [
            QtPdf.QPdfDocument.Status.Null,
            QtPdf.QPdfDocument.Status.Unloading,
        ]:
            self.current_file = filename
            self.document.load(filename)
        self.open_ext_button.setDisabled(False)
        self.search_prev_button.setDisabled(False)
        self.search_next_button.setDisabled(False)

    def goto(self, page):
        self.nav.jump(page - 1, QtCore.QPointF(0, 0))

    def search(self, str):
        self.model.setSearchString(str)
        self.view.setCurrentSearchResultIndex(0)
        self.view.viewport().repaint()
        if self.needstimer:
            self.numsearstart = time.perf_counter()
            self.numsearchTimer.start()
        self.update_search_label()
        self.search_label.setText(f"1 / {self.model.rowCount(QtCore.QModelIndex())}")

    def search_next(self, *, dir=1):
        if (
            self.view.sindex + dir < self.model.rowCount(QtCore.QModelIndex())
            and self.view.sindex + dir >= 0
        ):
            self.view.setCurrentSearchResultIndex(self.view.sindex + dir)
            self.update_search_label()

    def update_search_label(self):
        self.search_label.setText(
            f"{self.view.sindex+1} / {self.model.rowCount(QtCore.QModelIndex())}"
        )
        if self.needstimer and time.perf_counter() - self.numsearstart > 30:
            self.numsearchTimer.stop()

    # if (
    #     self.view.currentSearchResultIndex() + dir < self.model.rowCount(QtCore.QModelIndex())
    #     and self.view.currentSearchResultIndex() + dir >= 0
    # ):
    #     self.view.setCurrentSearchResultIndex(self.view.currentSearchResultIndex() + dir)
    #     self.search_label.setText(
    #         f"{self.view.currentSearchResultIndex()+1}"  # / {self.model.rowCount(QtCore.QModelIndex())}"
    #     )

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.document.close()
        return super().closeEvent(event)


# class Separator(QtWidgets.QWidget):
# def __init__(self):
# super().__init__()
# self.setFixedHeight(20)

# def paintEvent(self, event: QtGui.QPaintEvent) -> None:
# painter = QtGui.QPainter()
# painter.begin(self)
# painter.setRenderHint(QtGui.QPainter.Antialiasing)
# r = event.rect()
# r.adjust(30, 0, -30, 0)
# painter.fillRect(r, self.palette().base())
# painter.end()

try:
    text_icon = "/usr/share/icons/breeze-dark/mimetypes/64/text-plain.svg"
    if not os.path.isfile(text_icon):
        text_icon = (
            sp.check_output(["find", "/usr/share/icons", "-name", "text-plain.svg"])
            .decode()
            .split("\n")[0]
        )
except:
    text_icon = None
try:
    pdf_icon = "/usr/share/icons/breeze-dark/mimetypes/64/application-pdf.svg"
    if not os.path.isfile(pdf_icon):
        pdf_icon = (
            sp.check_output(["find", "/usr/share/icons", "-name", "application-pdf.svg"])
            .decode()
            .split("\n")[0]
        )
except:
    pdf_icon = None


class ResultWidget(LogEnabled, QtWidgets.QWidget):
    def __init__(self, parent, result: Result, pdf: RectPDF):
        super().__init__(parent)
        self.result = result
        self.pdf = pdf

        self.v_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.v_layout)
        self.v_layout.setSpacing(0)
        self.v_layout.setContentsMargins(0, 0, 0, 0)
        self.icon = QtWidgets.QWidget()
        w = Config.get("Results", "icon_width", int)
        self.icon.setFixedSize(w, w + 20)
        if self.result.type == "pdf":
            self.icon.setStyleSheet(f'image: url("{pdf_icon}");')
        elif self.result.type == "text":
            self.icon.setStyleSheet(f'image: url("{text_icon}");')
        elif self.result.type == "image":
            self.icon.setStyleSheet(f'image: url("{self.result.url}");')

        self.v_layout.addWidget(self.icon)
        self.v_layout.setAlignment(self.icon, QtCore.Qt.AlignmentFlag.AlignTop)
        self.layout = QtWidgets.QVBoxLayout()
        self.v_layout.addLayout(self.layout)
        self.filename = WrapLabel(self.result.filename)
        fft = Config.get("Optics.FontSizes", "filename_size", str)
        self.filename.setStyleSheet(f"font-size: {fft}pt;")
        self.location = WrapLabel(self.result.url)
        ht = self.result.highlighted_text.replace("[p ", "<br>[p ").strip()
        ht = ht.replace("[l ", "<br>[l ").strip()
        if ht.startswith("<br>"):
            ht = ht[4:]
        self.highlighted_text = QtWidgets.QLabel(ht)
        self.highlighted_text.setWordWrap(True)
        hft = Config.get("Optics.FontSizes", "context_size", str)
        self.highlighted_text.setStyleSheet(f"font-size: {hft}pt;")

        self.layout.addItem(QtWidgets.QSpacerItem(10, 10))
        self.layout.addWidget(self.filename)
        self.layout.addWidget(self.location)
        self.layout.addItem(QtWidgets.QSpacerItem(10, 5))
        self.layout.addWidget(self.highlighted_text)
        self.layout.addItem(QtWidgets.QSpacerItem(10, 10))

        palette = self.location.palette()
        color = palette.color(QtGui.QPalette.Text)
        color.setAlpha(127)
        palette.setColor(QtGui.QPalette.Text, color)
        self.location.setPalette(palette)
        self.layout.addStretch(1)
        self.show_line = Config.get("Optics", "show_list_separator", bool)
        self.line_thickness = Config.get("Optics", "list_separator_px", int)
        self.set_width(parent.width())

    def sizeHint(self):
        hint = super().sizeHint()
        # hint = self.geometry()
        return QtCore.QSize(hint.width(), hint.height() + 5 + self.line_thickness)

    def deactivate(self):
        self.location.hide()
        self.highlighted_text.hide()

    def activate(self):
        self.location.show()
        self.highlighted_text.show()

    def open_file(self):
        if self.result.url == "":
            return
        if self.result.type == "pdf" and self.result.url.lower().endswith("pdf"):
            self.log.debug(f"Opening PDF {self.result.url}")
            Config.text_doc.hide()
            Config.image_doc.hide()
            self.pdf.show()
            self.pdf.open(self.result.url)
            sn = [0, None, None] if len(self.result.snippets) == 0 else self.result.snippets[0]
            page, term, text = sn
            full_term = self.result.found_text
            self.pdf.goto(page)
            QtCore.QTimer.singleShot(15, lambda: self.pdf.search(full_term))
        elif self.result.type == "text":
            self.pdf.hide()
            Config.image_doc.hide()
            Config.text_doc.show()
            Config.text_doc.open(self.result)
        elif self.result.type == "image":
            self.pdf.hide()
            Config.text_doc.hide()
            Config.image_doc.show()
            Config.image_doc.setStyleSheet(f'image: url("{self.result.url}")')
        else:
            Config.image_doc.hide()
            self.pdf.hide()
            Config.text_doc.show()
            Config.text_doc.setText("Cannot display file")
        # else:
        # self.open_external()

    def open_external(self):
        if self.result.url == "":
            return
        sp.call(["xdg-open", self.result.url])

    def set_width(self, w):
        w = w - self.icon.width()
        self.highlighted_text.setFixedWidth(w)
        self.location.setFixedWidth(w)
        self.filename.setFixedWidth(w)
        self.highlighted_text.updateGeometry()
        self.location.updateGeometry()
        self.filename.updateGeometry()
        self.layout.invalidate()
        self.updateGeometry()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        super().paintEvent(event)
        if self.show_line:
            painter = QtGui.QPainter()
            painter.begin(self)
            painter.setRenderHint(QtGui.QPainter.Antialiasing)
            r = self.rect()
            rect = QtCore.QRect(
                0, r.bottom() - 1 - self.line_thickness, self.width(), 2 + self.line_thickness
            )
            painter.fillRect(rect, self.palette().base())
            rect = QtCore.QRect(
                int(self.width() * 0.1),
                r.bottom() - self.line_thickness,
                int(self.width() * 0.8),
                self.line_thickness,
            )
            painter.fillRect(rect, self.palette().mid())
            painter.end()


class ResultDisplay(LogEnabled, QtWidgets.QWidget):
    def __init__(self, pdf: RectPDF, tag_widget: TagWidget):
        super().__init__()
        self.pdf = pdf
        self.results = []
        self.search_results = []
        self.layout = QtWidgets.QVBoxLayout()
        self.list = QtWidgets.QListWidget()
        self.list.installEventFilter(self)
        self.list.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.list.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.list.setLayoutMode(QtWidgets.QListView.LayoutMode.Batched)
        self.layout.addWidget(self.list)
        self.setLayout(self.layout)
        self.list.currentItemChanged.connect(self.item_changed)
        self.list.currentItemChanged.connect(
            lambda item: tag_widget.change_file(self.list.itemWidget(item).result.url)
        )
        self.is_search_result = True

        self.open_timer = QtCore.QTimer()
        self.open_timer.setInterval(50)
        self.open_timer.setSingleShot(True)

        self.list.itemDoubleClicked.connect(self.open_external)
        palette = self.list.palette()
        linecol = palette.dark().color().name()
        base = palette.base().color().name()
        selected_bg = palette.highlight().color().name()
        selected_fg = palette.highlightedText().color().name()
        self.list.setStyleSheet(
            # f"QListWidget::item {{ border-bottom: 1px solid {linecol}; border-radius: 0px;}} "
            f"QListWidget::item {{ background-color: {base};}} "
            + f"QListWidget::item:selected {{ background-color: {selected_bg}; color: {selected_fg}; }}"
        )

        self.list.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.layout.setContentsMargins(0, 0, 0, 0)
        tag_widget.tag_filter_changed.connect(self.filter)
        tag_widget.tag_changed.connect(self.show_tag_files)
        self.tag_widget = tag_widget
        # self.list.setSpacing(2)
        # self.list.currentItemChanged.connect(self.cchanged)

    # def cchanged(self):
    #     self.list.itemWidget(self.selected).deactivate()
    #     self.selected.setSizeHint(self.list.itemWidget(self.selected).sizeHint())
    #     self.selected = self.list.currentItem()
    #     self.list.itemWidget(self.selected).activate()
    #     self.selected.setSizeHint(self.list.itemWidget(self.selected).sizeHint())

    def filter(self, list_of_files):
        status = self.parent().result_status.status()
        if len(list_of_files) == 0:
            for i in range(self.list.count()):
                self.list.item(i).setHidden(False)
            self.parent().result_status.setStatus(status & ~ResultStatus.State.FILTERED)
        else:
            first_found = False
            for i in range(self.list.count()):
                item = self.list.item(i)
                if self.list.itemWidget(item).result.url in list_of_files:
                    item.setHidden(False)
                    if not first_found:
                        self.list.setCurrentRow(i)
                        first_found = True
                else:
                    item.setHidden(True)
            self.parent().result_status.setStatus(status | ResultStatus.State.FILTERED)

    def item_changed(self, i):
        if self.open_timer.isActive():
            self.open_timer.stop()
            self.open_timer.timeout.disconnect()
        self.open_timer.timeout.connect(lambda i=i: self.open_file(i))
        self.open_timer.start()

    def open_file(self, item):
        self.log.debug("Telling item to open")
        self.list.itemWidget(item).open_file()
        try:
            self.open_timer.timeout.disconnect()
        except:
            self.log.debug("was not connected")

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self.open_timer.isActive():
            self.open_timer.stop()
            self.open_timer.timeout.disconnect()
        return super().closeEvent(event)

    def open_external(self, item: QtWidgets.QListWidgetItem):
        self.list.itemWidget(item).open_external()

    def eventFilter(self, watched, event):
        if event.type() == QtCore.QEvent.KeyPress:
            if event.matches(QtGui.QKeySequence.InsertParagraphSeparator):
                self.list.itemWidget(self.list.currentItem()).open_external()
            elif event.key() == QtCore.Qt.Key_Space:
                self.list.itemWidget(self.list.currentItem()).open_file()
            elif (
                event.key() == QtCore.Qt.Key_C
                and QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
            ):
                cb = QtWidgets.QApplication.clipboard()
                url = self.list.itemWidget(self.list.currentItem()).result.url
                cb.setText(url)
        return False

    def show_tag_files(self, files):
        tag, files = files
        if len(files) > 0 and files[0] == "":
            if Config.get("Results", "revert_to_search_results", bool):
                if self.is_search_result:
                    return
                r = self.search_results
                tags = False
            else:
                return
        else:
            self.parent().result_status.setStatus(ResultStatus.State.TAGS)
            self.parent().search_input.setText(f"[{tag}]" if tag != "" else "")
            r = [
                Result(f, os.path.basename(f), "", "", "", "", "", "", check_mime(f)) for f in files
            ]
            tags = True
        self.show_results(r, show_tags=tags)

    def show_results(self, results: typing.List[Result], show_tags=False):
        if self.open_timer.isActive():
            self.open_timer.stop()
            self.open_timer.timeout.disconnect()

        self.is_search_result = not show_tags
        if len(results) == 0:
            results = [Result("", "No Results", "", "", "", "", "", "", "")]
        self.list.blockSignals(True)
        self.results = results
        if self.is_search_result:
            self.search_results = results
        self.list.clear()
        # Weird magical numbers (2* and 6) but I don't know how to effectively get the internal width of a QListWidget
        content_width = self.list.width() - 3 * self.list.verticalScrollBar().width() - 6
        for r in results:
            w = ResultWidget(self.list, r, self.pdf)
            it = QtWidgets.QListWidgetItem()
            self.list.addItem(it)
            self.list.setItemWidget(it, w)
            w.set_width(content_width)
            it.setSizeHint(w.sizeHint())
        self.list.blockSignals(False)
        if results[0].url != "":
            self.list.setCurrentRow(0)
        if self.is_search_result:
            self.parent().result_status.setStatus(ResultStatus.State.SEARCH)
        # self.tag_widget.update_list_of_files(self.search_results)
        self.tag_widget.update_list_of_files(results)

    def resizeEvent(self, ev):
        content_width = self.list.width() - 2 * self.list.verticalScrollBar().width() - 6
        if content_width < 350:
            content_width = 350
            self.list.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        else:
            self.list.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        for i in range(self.list.count()):
            w = self.list.itemWidget(self.list.item(i))
            w.set_width(content_width)
            self.list.item(i).setSizeHint(w.sizeHint())
        super().resizeEvent(ev)


class EnterLineEdit(QtWidgets.QLineEdit):
    enterPressed = QtCore.Signal()

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.matches(QtGui.QKeySequence.InsertParagraphSeparator):
            self.enterPressed.emit()
        return super().keyPressEvent(event)


class Overlay(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.text = ""
        self.color = QtGui.QColor(0, 0, 0, 125)

    def show(self, x, y, width, height, text, color=(0, 0, 0, 125)):
        self.text = text
        self.move(x, y)
        self.resize(width, height)
        self.color = QtGui.QColor(*color)
        super().show()

    def paintEvent(self, event):
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.fillRect(event.rect(), QtGui.QBrush(self.color))
        font = painter.font()
        font.setPointSize(18)
        font.setWeight(QtGui.QFont.ExtraBold)
        painter.setFont(font)
        painter.setPen(QtGui.QColor(255, 255, 255))
        painter.drawText(self.rect(), QtCore.Qt.AlignCenter, self.text)
        painter.end()


class DisableInputFilter(QtCore.QObject):
    def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent) -> bool:
        ev = QtCore.QEvent
        if event.type() in [
            ev.KeyPress,
            ev.KeyRelease,
            ev.MouseButtonPress,
            ev.MouseButtonDblClick,
            ev.MouseMove,
            ev.HoverEnter,
            ev.HoverLeave,
            ev.HoverMove,
            ev.DragEnter,
            ev.DragLeave,
            ev.DragMove,
            ev.Drop,
        ]:
            return True

        return super().eventFilter(watched, event)


class LED(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        # self.setFixedSize(20, 20)
        self.state = True
        if self.palette().dark().color().lightness() < self.palette().base().color().lightness():
            self.off_col = self.palette().dark()
        else:
            self.off_col = self.palette().base()
        self.on_col = self.palette().highlight()
        self.border = self.palette().mid().color()

    def sizeHint(self):
        return QtCore.QSize(17, 17)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        if self.state:
            painter.setBrush(self.on_col)
        else:
            painter.setBrush(self.off_col)
        painter.setPen(self.border)
        # painter.setPen(painter.brush().color())
        s = min(self.height() - 2, self.width() - 2)
        ellipse_xstart = (self.width() - 2 - s) / 2 + 1
        ellipse_ystart = (self.height() - 2 - s) / 2 + 1
        painter.drawEllipse(ellipse_xstart, ellipse_ystart, s, s)
        painter.end()
        return super().paintEvent(event)


class ResultStatus(QtWidgets.QWidget):
    class State(enum.Flag):
        NOTHING = enum.auto()
        SEARCH = enum.auto()
        TAGS = enum.auto()
        FILTERED = enum.auto()

    def __init__(self):
        super().__init__()
        self._status = self.State.NOTHING
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(QtWidgets.QLabel("Results are:"))
        self.led_search = LED()
        self.layout.addWidget(self.led_search)
        self.layout.addWidget(QtWidgets.QLabel("Search"))
        self.led_tags = LED()
        self.layout.addWidget(self.led_tags)
        self.layout.addWidget(QtWidgets.QLabel("Tags"))
        self.led_filtered = LED()
        self.layout.addWidget(self.led_filtered)
        self.layout.addWidget(QtWidgets.QLabel("Filtered"))
        self.led_tags.state = False
        self.layout.addStretch(1)

    def setStatus(self, state):
        self._status = state
        self.led_search.state = state & self.State.SEARCH
        self.led_tags.state = state & self.State.TAGS
        self.led_filtered.state = state & self.State.FILTERED
        self.led_search.repaint()
        self.led_tags.repaint()
        self.led_filtered.repaint()

    def status(self):
        return self._status


class SearchWidget(LogEnabled, QtWidgets.QWidget):
    def __init__(self, parent, pdf, pdfgrep_dir, search_term, max_found, tag_widget):
        super().__init__(parent)
        self.pdf = pdf
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.search_input = EnterLineEdit()
        self.search_input.setMinimumHeight(40)
        QtCore.QTimer.singleShot(10, self.search_input.setFocus)
        f = self.search_input.font()
        f.setPointSize(f.pointSize() + 3)
        # f.setBold(True)
        self.search_input.setFont(f)
        self.search_input.setPlaceholderText("Search ...")
        # c = self.search_input.palette().base().color().name()
        self.search_input.setStyleSheet("padding: 10px;")
        # f"background-color: {c}; border-radius: 15px; padding: 10px;"
        # )
        self.search_button = QtWidgets.QPushButton("Search")
        self.search_button.setMinimumHeight(40)
        self.search_layout = QtWidgets.QHBoxLayout()
        self.search_layout.addWidget(self.search_input)
        self.search_layout.addWidget(self.search_button)
        self.search_label = QtWidgets.QLabel("")
        self.layout.addLayout(self.search_layout)
        self.result_status = ResultStatus()
        self.result_status.setVisible(Config.get("Results", "show_result_status", bool))
        # self.layout.addWidget(self.search_label)
        self.layout.addWidget(self.result_status)
        self.search_button.clicked.connect(self.search)
        self.result_display = ResultDisplay(self.pdf, tag_widget)
        self.tag_widget = tag_widget
        self.layout.addWidget(self.result_display)

        if max_found is None:
            max_found = Config.get("Results", "limit_num", int)
        self.pdfgrep_dir = pdfgrep_dir
        self.max_found = max_found
        self.engine = None

        self.search_input.enterPressed.connect(self.search)
        self.overlay = Overlay(self.parent())
        self.overlay.hide()
        self.input_blocker = DisableInputFilter()
        if search_term is not None:
            self.search_input.setText(search_term)
            QtCore.QTimer.singleShot(100, self.search)

    def engines_changed(self, engine_type):
        if not isinstance(engine_type, (tuple, list)):
            engine_type = [engine_type]
        self.engine = Engine.create(
            engine_type,
            path=self.pdfgrep_dir,
            max_found=self.max_found,
            timeout=Config.get("Search", "timeout", int),
        )
        self.engine.search_finished.connect(self.search_finished)

    def add_all_tags(self, tag):
        for r in self.result_display.results:
            self.tag_widget.tagger.add_tags_to_file(r.url, tag)
        self.tag_widget.tagger.save()
        self.tag_widget.list.model().layoutChanged.emit()

    def search(self):
        Config.text_doc.setText("")
        Config.image_doc.setStyleSheet("")
        self.pdf.document.close()
        if self.engine is None:
            self.engines_changed(Config.get("Search", "engines"))
        text = self.search_input.text()
        self.result_display.list.setFocus()
        self.tag_widget.clear_selection()
        if text == "":
            self.tag_widget.change_file(None)
            self.result_display.show_results([], show_tags=True)
            self.result_status.setStatus(ResultStatus.State.TAGS)
            return
        elif text.startswith("[") and text.endswith("]"):
            self.log.debug("Searched for tag")
            self.tag_widget.search_tag(text[1:-1])
            return
        self.overlay.show(
            0,
            0,
            self.parent().parent().parent().width(),
            self.parent().parent().parent().height(),
            "Searching ...",
        )
        QtWidgets.QApplication.instance().installEventFilter(self.input_blocker)
        results = self.engine.search(text)

    def search_finished(self, results):
        self.overlay.hide()
        QtWidgets.QApplication.instance().removeEventFilter(self.input_blocker)
        self.result_display.show_results(results)
        QtCore.QTimer.singleShot(10, self.result_display.list.setFocus)


class TextDoc(LogEnabled, QtWidgets.QTextEdit):
    def __init__(self):
        super().__init__()
        self.setReadOnly(True)
        self.text = ""
        self.cpos = 0
        self.result = None
        bg = Config.get("Optics", "text_highlight_bg_color", str)
        fg = Config.get("Optics", "text_highlight_fg_color", str)
        self.setStyleSheet(
            f"TextDoc{{selection-background-color: '{bg}';selection-color: '{fg}';}}"
        )

    def open(self, result):
        self.result = result
        try:
            with open(result.url, "r") as f:
                self.text = f.read()
                self.setText(self.text.lower())
            ln = result.snippets[0][0]
            if ln is not None and ln != 0:
                doc = self.document()
                cursor = QtGui.QTextCursor(doc.findBlockByLineNumber(ln - 1))
                pos = cursor.position()
                self.cpos = pos
                # cursor.select(QtGui.QTextCursor.SelectionType.LineUnderCursor)
                # text = cursor.selectedText()
                # start = text.index(result.search_text)
                # cursor.setPosition(pos + start)
                # cursor.setPosition(
                #     pos + start + len(result.search_text),
                #     QtGui.QTextCursor.MoveMode.KeepAnchor,
                # )
                # self.setTextCursor(cursor)
                # self.ensureCursorVisible()
                self.search_next(1)
        except Exception as e:
            self.log.error(e)
            self.setText("Cannot display file")

    def search_next(self, dir=1):
        if dir == 0:
            return
        if dir > 0:
            pos = self.text.find(self.result.search_text, self.cpos + len(self.result.search_text))
        else:
            pos = self.text.rfind(self.result.search_text, 0, self.cpos)

        if pos < 0:
            return

        doc = self.document()
        cursor = self.textCursor()
        cursor.setPosition(pos)
        cursor.setPosition(
            pos + len(self.result.search_text),
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        self.setTextCursor(cursor)
        self.ensureCursorVisible()
        self.cpos = pos


class DirectorySelector(QtWidgets.QTreeView):
    dir_changed = QtCore.Signal(str)

    def __init__(self, dir):
        super().__init__()
        self.model = QtWidgets.QFileSystemModel()
        self.model.setRootPath(os.path.expanduser("~"))
        self.setModel(self.model)
        self.setRootIndex(self.model.index(self.model.rootPath()))
        [self.hideColumn(i) for i in range(1, 5)]
        self.model.setFilter(QtCore.QDir.Filter.Dirs)
        self.change_dir(dir)
        self.blockSignals(False)
        self.doubleClicked.connect(self.doubleclick)
        self.clicked.connect(self.singleclicked)

    def singleclicked(self, index):
        self.setExpanded(index, not self.isExpanded(index))

    def doubleclick(self, index):
        self.setRootIndex(self.model.index(self.model.filePath(index)))

    def currentChanged(self, index, previous):
        path = self.model.filePath(index)
        if path == "":
            return
        self.dir_changed.emit(path)

    def change_dir(self, dir):
        self.selectionModel().clearSelection()
        # path = dir
        # while path != "" and path != "/":
        # path, _ = os.path.split(path)
        # self.expand(self.model.index(path))
        idx = self.model.index(dir)
        self.selectionModel().select(idx, QtCore.QItemSelectionModel.Select)
        self.scrollTo(idx, QtWidgets.QAbstractItemView.ScrollHint.PositionAtCenter)
        QtWidgets.QApplication.processEvents()
        self.scrollTo(idx, QtWidgets.QAbstractItemView.ScrollHint.PositionAtCenter)


class MainWidget(QtWidgets.QWidget):
    def __init__(self, parent, pdfgrep_dir, search_term, initial_zoom, max_found):
        super().__init__(parent)
        # self.setLayout(QtWidgets.QHBoxLayout())
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.splitter = QtWidgets.QSplitter()
        self.layout.addWidget(self.splitter)
        self.pdf = PDF(initial_zoom)
        self.text_doc = TextDoc()
        self.text_doc.hide()
        Config.text_doc = self.text_doc
        self.image_doc = QtWidgets.QWidget()
        Config.image_doc = self.image_doc
        self.tagger = Tagger(Config.get("Results", "tag_database_path"))
        self.tag_widget = TagWidget(
            self, self.tagger, Config.get("Results", "comments_as_markdown", bool)
        )
        self.search = SearchWidget(
            self, self.pdf, pdfgrep_dir, search_term, max_found, self.tag_widget
        )
        self.dir_sel = DirectorySelector(pdfgrep_dir)
        self.dir_sel.dir_changed.connect(self.path_changed)
        self.leftsplit = QtWidgets.QSplitter()
        self.leftsplit.addWidget(self.dir_sel)
        self.leftsplit.addWidget(self.search)
        large_number = int(1e6)
        if Config.get("Optics", "dirtree_shown", bool):
            w = Config.get("Optics", "dirtree_width_percent", int) / 100
            self.leftsplit.setSizes(
                [
                    int(w * large_number),
                    int((1 - w) * large_number),
                ]
            )
        else:
            self.leftsplit.setSizes([0, large_number])
        self.splitter.addWidget(self.leftsplit)
        containerwidget = QtWidgets.QWidget()
        containerwidget.setLayout(QtWidgets.QHBoxLayout())
        containerwidget.layout().addWidget(self.pdf)
        containerwidget.layout().addWidget(self.text_doc)
        containerwidget.layout().addWidget(self.image_doc)
        containerwidget.layout().setSpacing(0)
        containerwidget.setMinimumWidth(300)
        # containerwidget.layout().setContentsMargins(0, 0, 0, 0)
        # self.splitter.addWidget(self.pdf)
        # self.splitter.addWidget(containerwidget)
        self.rightsplitter = QtWidgets.QSplitter()
        self.rightsplitter.addWidget(containerwidget)
        self.rightsplitter.addWidget(self.tag_widget)
        self.rightsplitter.setSizes([1, int(Config.get("Optics", "tags_shown", bool))])

        if Config.get("Optics", "tags_shown", bool):
            w = Config.get("Optics", "tags_width_percent", int) / 100
            self.rightsplitter.setSizes(
                [
                    int((1 - w) * large_number),
                    int(w * large_number),
                ]
            )
        else:
            self.rightsplitter.setSizes([large_number, 0])

        self.splitter.addWidget(self.rightsplitter)
        w = Config.get("Optics", "viewer_width_percent", int) / 100
        self.splitter.setSizes(
            [
                int((1 - w) * large_number),
                int(w * large_number),
            ]
        )
        self.search.setFocus()
        self.search.search_input.setFocus()
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def show_all_tags(self):
        # self.search.search_input.setText("")
        # self.search.search_input.enterPressed.emit()
        self.tag_widget.switch_to_all_tags()
        self.show_tags()

    def add_all_tags(self):
        tag, ok = QtWidgets.QInputDialog.getText(
            self, "Enter tag to apply", "Enter the tag name to apply to all search results"
        )
        if ok:
            self.search.add_all_tags(tag)

    def show_tags(self, toggle=False):
        large_number = int(1e6)
        if toggle:
            m = self.rightsplitter.sizes()[1] == 0
        else:
            m = 1
        w = Config.get("Optics", "tags_width_percent", int) / 100
        self.rightsplitter.setSizes(
            [
                int((1 - w) * large_number),
                int(w * large_number) * m,
            ]
        )

    def toggle_dir_sel(self):
        sizes = self.leftsplit.sizes()
        large_number = int(1e6)
        if sizes[0] == 0:
            w = Config.get("Optics", "dirtree_width_percent", int) / 100
            self.leftsplit.setSizes(
                [
                    int(w * large_number),
                    int((1 - w) * large_number),
                ]
            )
        else:
            self.leftsplit.setSizes([0, large_number])

    def path_changed(self, path):
        self.search.pdfgrep_dir = path
        self.search.engines_changed(self.parent().enabled_engines)

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key_F3:
            if self.pdf.isVisible():
                self.pdf.search_next(
                    dir=(
                        -1
                        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier
                        else 1
                    )
                )
            elif self.text_doc.isVisible():
                self.text_doc.search_next(
                    dir=(
                        -1
                        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier
                        else 1
                    )
                )

            # self.pdf.search_next()
        return super().keyPressEvent(event)


class NumResDialog(QtWidgets.QDialog):
    def __init__(self, parent, val):
        super().__init__(parent)
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(QtWidgets.QLabel("Maximum Number of Results to Display:"))
        self.input = QtWidgets.QSpinBox()
        self.input.setMaximum(1000000)
        self.input.setValue(val)
        self.layout.addWidget(self.input)
        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )
        self.layout.addWidget(self.buttons)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)


class InitialDisplay(QtWidgets.QWidget):
    do_search = QtCore.Signal(str)

    def __init__(self, parent, width, height):
        super().__init__(parent)
        self.initial_input = EnterLineEdit(self)
        f = self.initial_input.font()
        f.setPointSize(f.pointSize() + 3)
        self.initial_input.setFont(f)

        self.resize(width, height)
        self.initial_input.setFixedWidth(0.5 * self.width())
        self.initial_input.setFixedHeight(40)
        # self.initial_input.move(0.1 * self.width(), 0.3 * self.height())

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.addStretch(1)
        l1 = QtWidgets.QLabel("PDFSearch")
        l1.setStyleSheet("font-size: 20pt")
        self.layout.addWidget(l1)
        self.layout.addStretch(1)
        l2 = QtWidgets.QLabel("Press Enter to Search")
        l2.setStyleSheet("font-size: 13pt")
        self.layout.addWidget(l2)
        self.layout.addWidget(self.initial_input)
        self.layout.addStretch(2)
        self.layout.setAlignment(self.initial_input, QtCore.Qt.AlignCenter)
        self.layout.setAlignment(l1, QtCore.Qt.AlignCenter)
        self.layout.setAlignment(l2, QtCore.Qt.AlignCenter)

        self.initial_input.enterPressed.connect(
            lambda: self.parent().initial_done(self.initial_input.text())
        )
        self.setAutoFillBackground(True)
        self.initial_input.setFocus()

    def hide(self):
        super().hide()
        # QtCore.QTimer.singleShot(100, setup)


class MainWindow(QtWidgets.QMainWindow):
    engines_changed = QtCore.Signal(list)

    def __init__(self, engine, pdfgrep_dir, search_term, initial_zoom, max_found):
        super().__init__()
        self.widget = MainWidget(self, pdfgrep_dir, search_term, initial_zoom, max_found)
        self.setCentralWidget(self.widget)

        self.resize(1600, 800)

        self.setWindowTitle("PDF Search")
        style = self.style()
        icon = style.standardIcon(style.StandardPixmap.SP_FileDialogContentsView)
        self.setWindowIcon(icon)
        self.toolbar = self.addToolBar("Engines")
        self.toolbar.addWidget(QtWidgets.QLabel("Search Engines: "))

        def ensure_list(c):
            if not isinstance(c, (tuple, list)):
                return [c]
            else:
                return c

        self.enabled_engines = []
        if engine is not None:
            self.enabled_engines = ensure_list(engine)
            self.enabled_engines = [
                i for i in self.enabled_engines if f"{i}search" in Engine.available_engines.keys()
            ]
        if len(self.enabled_engines) == 0:
            self.enabled_engines: list = ensure_list(Config.get("Search", "engines"))

        for engine in Engine.available_engines.values():
            action = self.toolbar.addAction(engine.name)
            action.setCheckable(True)
            action.setChecked(engine.name.lower() in self.enabled_engines)
            action.triggered.connect(
                lambda x, engine=engine.name.lower(): self.update_engines(x, engine)
            )

        self.engines_changed.connect(self.widget.search.engines_changed)
        self.engines_changed.emit(self.enabled_engines)
        self.toolbar.addSeparator()
        act = self.toolbar.addAction("Show All Tags")
        act.triggered.connect(self.widget.show_all_tags)

        self.file_menu = self.menuBar().addMenu("File")
        q = self.file_menu.addAction("Quit")
        q.triggered.connect(self.close)
        act = self.file_menu.addAction("Change Directory")
        act.triggered.connect(self.change_directory)
        act.setShortcut("Ctrl+D")
        act = self.file_menu.addAction("Open Configuration File")
        act.triggered.connect(lambda: sp.run(["xdg-open", Config.file]))
        self.view_menu = self.menuBar().addMenu("View")
        act = self.view_menu.addAction("Toggle Directory Chooser")
        act.triggered.connect(self.widget.toggle_dir_sel)
        act.setShortcut("F2")
        act = self.view_menu.addAction("Toggle Tag List")
        act.triggered.connect(lambda: self.widget.show_tags(toggle=True))
        act.setShortcut("F4")

        self.numres_dialog = None

        self.search_menu = self.menuBar().addMenu("Search")
        act = self.search_menu.addAction("Change Max Number of Results")
        act.triggered.connect(self.show_change_numres_dialog)
        act = self.search_menu.addAction("Show All Tags")
        act.triggered.connect(self.widget.show_all_tags)
        act.setShortcut("Ctrl+S")
        act = self.search_menu.addAction("Add Tag For All Results")
        act.triggered.connect(self.widget.add_all_tags)
        act.setShortcut("Ctrl+Shift+T")

        self.help_menu = self.menuBar().addMenu("Help")
        act = self.help_menu.addAction("Show Help")
        act.triggered.connect(self.show_help)
        about = self.help_menu.addAction("About")
        about.triggered.connect(self.show_about)
        about = self.help_menu.addAction("About Qt")
        about.triggered.connect(lambda: QtWidgets.QMessageBox.aboutQt(self))
        # self.initial_widget = InitialDisplay(self, self.width(), self.height())
        # self.initial_widget.do_search.connect(self.initial_done)
        # self.initial_widget.show()

    def initial_done(self, text):
        self.widget.search.search_input.setText(text)
        QtWidgets.QApplication.processEvents()
        self.initial_widget.hide()
        self.initial_widget.do_search.disconnect()
        self.initial_widget.setParent(None)
        QtCore.QTimer.singleShot(10, self.widget.search.search_input.enterPressed.emit)

    def show_change_numres_dialog(self):
        if self.numres_dialog is None:
            self.numres_dialog = NumResDialog(self, self.widget.search.max_found)
        if self.numres_dialog.exec():
            self.widget.search.max_found = self.numres_dialog.input.value()
            self.widget.search.engines_changed(self.enabled_engines)

    def show_about(self):
        QtWidgets.QMessageBox.about(
            self,
            "PDFSearch",
            """
PDFSearch is a simple gui frontend for various search providers on linux\n
Copyright: Patrick Schreiber
        """,
        )

    def show_help(self):
        self.help = QtWidgets.QTextEdit()
        self.help.setReadOnly(True)
        self.help.resize(500, 500)
        self.help.setMarkdown(
            """
# PDFSearch

PDFsearch is a simple gui frontend to various search providers:

- Recoll
- pdfgrep
- ag (silver-searcher)
- find

The providers can be combined which will produce a list with the results of each
provider after the other. To select/deselect providers, click on their name at the top of the window.

For all but recoll, the search path can be choosen, either in the tree (View->Toggle Directory Chooser)
or with "File->ChangeDirectory".

For recoll the directory depends on the default index of recoll. To set this, use the official recoll gui
and refer to the recoll documentation.
        """
        )
        self.help.show()

    def change_directory(self):
        dir = QtWidgets.QFileDialog.getExistingDirectory(self)
        self.widget.dir_sel.change_dir(dir)
        # self.widget.search.pdfgrep_dir = dir
        # self.engines_changed.emit(self.enabled_engines)

    def update_engines(self, state, engine):
        if state:
            if engine not in self.enabled_engines:
                self.enabled_engines.append(engine)
        else:
            if engine in self.enabled_engines:
                del self.enabled_engines[self.enabled_engines.index(engine)]
        self.engines_changed.emit(self.enabled_engines)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.widget.pdf.close()
        if self.widget.search.engine is not None:
            self.widget.search.engine.stop()
        super().closeEvent(event)


def main():
    import argparse as ap

    config_file_path = os.path.join(os.path.expanduser("~"), ".config", "pdfsearch.cfg")

    parser = ap.ArgumentParser("pdfsearch", description="PDF Search - GUI for pdfgrep and recoll")
    parser.add_argument(
        "--engine",
        default=None,
        help="Select search engine to use",
        nargs="+",
        # choices=["pdfgrep", "recoll", "ag", "ag+pdfgrep", "pdfgrep+ag"],
    )
    parser.add_argument(
        "--path",
        help="Set the search path for pdfgrep, no effect with recoll",
    )
    parser.add_argument(
        "--initial-zoom", "-z", type=int, default=100, help="Initial zoom of the pdf window"
    )
    parser.add_argument(
        "--limit-num",
        "-n",
        type=int,
        default=None,
        help="Limit number of results to this. This speeds up the displaying of results. "
        "With pdfgrep search, the limit applies to content and filenames individually.",
    )
    parser.add_argument(
        "--config-file",
        default=config_file_path,
        help=f"Set the config file path. Default: {config_file_path}",
    )
    parser.add_argument(
        "--write-default-config",
        action="store_true",
        help="Write the default configuration to the given (or default) config file if the file is empty or does not exist",
    )
    parser.add_argument("initial", nargs="*", help="Set the initial search term")
    parser.add_argument("--log", help="Set the log level")
    args = parser.parse_args()

    initial = " ".join(args.initial) if len(args.initial) > 0 else None

    Config.read(args.config_file)
    if args.write_default_config:
        if os.path.isfile(args.config_file):
            with open(args.config_file, "r") as f:
                if len(f.read()) > 0:
                    log.debug("Config file not empty")
                    exit(1)
        with open(args.config_file, "w") as f:
            Config._config.write(f)
        exit(0)

    # if args.engine is not None:
    # Config.set("Search", "engines", ",".join(args.engine))

    if args.path is None:
        args.path = Config.get("Search", "path")

    if args.log:
        logging.basicConfig(
            level=getattr(logging, args.log.upper()),
            format='%(asctime)s %(levelname)s %(name)-15s: %(message)s',
        )

    # QtWidgets.QApplication.setDesktopSettingsAware(False)
    app = QtWidgets.QApplication(["PDF Search"])
    # palette = QtGui.QPalette()
    # palette.setColor(QtGui.QPalette.Window, QtGui.QColor(3, 18, 14))
    # palette.setColor(QtGui.QPalette.Base, QtGui.QColor(15, 15, 15))
    # palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.ToolTipBase, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.Text, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.BrightText, QtCore.Qt.red)
    # palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(142, 45, 197).lighter())
    # palette.setColor(QtGui.QPalette.HighlightedText, QtCore.Qt.black)
    # app.setPalette(palette)
    mw = MainWindow(args.engine, args.path, initial, args.initial_zoom, args.limit_num)
    mw.show()
    app.exec_()

    # TODO:
    # - Add SearchEngine baseclass
    # - Add search finished signal to engine
    # - Put actual search in a background thread
    # - Emit search finished signal when finished searching
    # - Connect search finished signal to show_results


if __name__ == "__main__":
    main()
