from qtpy import QtWidgets, QtCore, QtGui
import enum


class Switch(QtWidgets.QWidget):

    class State(enum.Flag):
        Left = True
        Right = False

    stateChanged = QtCore.Signal(State)

    def __init__(self, left_text, right_text, start_state: State = State.Left):
        super().__init__()
        self.left_text = left_text
        self.right_text = right_text
        self.state = start_state
        self.setMinimumHeight(20)
        self._height_frac = 0.6
        self.margin = 10
        self.max_center_width = 1.7
        metrics = QtGui.QFontMetrics(self.font())
        self.left_size = metrics.size(QtCore.Qt.TextFlag.TextSingleLine, self.left_text)
        self.right_size = metrics.size(QtCore.Qt.TextFlag.TextSingleLine, self.right_text)
        self.left_size.setWidth(self.left_size.width() + self.margin)
        self.right_size.setWidth(self.right_size.width() + self.margin)
        self.setMinimumWidth(
            self.left_size.width() + self.right_size.width() + 30
        )  # 40 = 2*minHeight

        palette = self.palette()
        if palette.dark().color().lightness() < palette.base().color().lightness():
            self.track_col = palette.dark().color()
        else:
            self.track_col = palette.base().color()
        self.text_col = palette.text().color()
        self.dot_col = palette.highlight().color()

    def setState(self, state):
        old_state = self.state
        self.state = state
        self.repaint()
        if state != old_state:
            self.stateChanged.emit(state)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        # self.state = not self.state
        # self.stateChanged.emit(self.state)
        # self.repaint()
        self.setState(~self.state)
        return super().mouseReleaseEvent(event)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        super().paintEvent(event)
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        rect = self.rect()

        painter.setPen(self.text_col)
        painter.drawText(
            0,
            0,
            self.left_size.width(),
            rect.height(),
            QtCore.Qt.AlignmentFlag.AlignCenter,
            self.left_text,
        )

        painter.setBrush(self.track_col)
        painter.setPen(self.track_col)
        available_width = min(
            rect.width() - self.left_size.width() - self.right_size.width(),
            (self.max_center_width * self.height()),
        )
        dot_height = min((rect.height() - 2), int(available_width / 2))
        dot_ypos = (rect.height() - dot_height) / 2
        # m = int(dot_height * ((1 - (self._height_perc / 100))) / 2)
        bar_height = int(dot_height * self._height_frac)
        bar_ypos = dot_ypos + (dot_height - bar_height) / 2
        # h = int((rect.height() - 2 * m) / 2)
        painter.drawRoundedRect(
            self.left_size.width() + 1,
            bar_ypos,
            available_width - 2,
            bar_height,
            int(bar_height / 2),
            int(bar_height / 2),
        )
        painter.setBrush(self.dot_col)
        painter.setPen(self.dot_col)
        if self.state == self.State.Left:
            lpos = self.left_size.width()
        else:
            lpos = self.left_size.width() + available_width - dot_height
        painter.drawEllipse(
            lpos,
            dot_ypos,
            # rect.height() - 2,
            # rect.height() - 2,
            dot_height,
            dot_height,
        )
        painter.setPen(self.text_col)
        painter.drawText(
            self.left_size.width() + available_width,
            0,
            self.right_size.width(),
            rect.height(),
            QtCore.Qt.AlignmentFlag.AlignCenter,
            self.right_text,
        )
        painter.end()
