from qtpy import QtWidgets, QtGui, QtCore
import json
import hashlib
import os
import typing
from .switch import Switch
import logging
import time
from .log import LogEnabled

log = logging.getLogger("PDFSearch.tagging")

"""
JsonStructure:
    {
        'filehashes': {
            <filepath>: <filehash>
        },
        'files': {
            <filehash>: {
                'filename': <filename>,
                'hash': <hash>,
                'tags': [list of tag ids]
                'comments': ""
            }
        },
        'tags': {
            <tagid>: {
                'name': <tagname>,
                'files': [list of filehashes]
            }
        }
    }
"""
minimal_structure = {"filehashes": {}, "files": {}, "tags": {}}


class Tagger:
    def __init__(self, db_file: str):
        self.db_file = db_file
        if not os.path.exists(db_file):
            with open(db_file, "w") as f:
                json.dump(minimal_structure, f)
        with open(db_file, "r") as f:
            self._db = json.load(f)
        self._files = self._db["files"]
        self._hashes = self._db["filehashes"]
        self._tags = self._db["tags"]
        # self._tags = {}
        # for id, t in self._db["tags"]:
        # self._tags[int(id)] = t
        # self._db["tags"] = self._tags
        self._create_name_to_id_mapping()

    def number_of_tags(self) -> int:
        """Get the number of tags"""
        return len(self._tags)

    def available_tags(
        self, filecount: bool = False
    ) -> typing.List[typing.Union[str, typing.List[str | int]]]:
        """
        Get a list of existing tags

        Args:
            filecount: True to return a list of [tagnames, filecount_per_tag]
        Returns:
            List of existing tags
        """
        if not filecount:
            return [t["name"] for t in self._tags.values()]
        else:
            r = []
            for t in self._tags.values():
                r.append([t["name"], len(t["files"])])
            return r

    def set_comment(self, filename: str, comment: str):
        """
        Set a comment for a file

        Args:
            filename: The filename
            comment: The comment
        """
        hash, fobj = self.get_or_create_file(filename)
        fobj["comments"] = comment

    def get_comment(self, filename: str) -> str:
        """
        Get the comment for a file

        Args:
            filename: The filename
        Returns:
            The associated comment
        """
        hash, fobj = self.get_or_create_file(filename, store=False)
        return fobj["comments"]

    def _create_name_to_id_mapping(self):
        """Create the dictionary that maps tagnames to tagids"""
        self._tagids = {}
        for i, t in self._tags.items():
            self._tagids[t["name"]] = i

    def rename_tag(self, tagname: str, newtagname: str):
        """
        Change the name of a tag

        Args:
            tagname: The old tag name
            newtagname: The new tag name
        """
        if tagname == newtagname:
            return
        id = self.tag_id_from_name(tagname)
        if len(id) == 0:
            return
        else:
            id = id[0]
        if newtagname in self._tagids:
            # Tag name exists already, two possibilities, we combine them or we disalow the renaming
            # for now we combine them
            for hash in self._tags[id]["files"]:
                filename = self._files[hash]["filename"]
                self.remove_tags_from_file(filename, tagname)
                self.add_tags_to_file(filename, newtagname)
                # if new_id not in self._files[hash]["tags"]:
                # self._files[hash]["tags"].append(new_id)
                # if id in self._files[hash]["tags"]:
                # del self._files[hash]["tags"][self._files[hash]["tags"].index(id)]
            self.delete_tag(tagname)
        else:
            self._tags[id]["name"] = newtagname
        self._create_name_to_id_mapping()

    def compute_hash(self, filename: str) -> str:
        """
        Compute the hash of a file used as "id" in the database

        Args:
            filename: The path to the file
        Returns:
            The hash of the first 1024 bytes of the file as hexdigest
        """
        try:
            with open(filename, "rb") as f:
                h = hashlib.md5()
                h.update(f.read(1024))
                return h.hexdigest()
        except:
            return None

    def create_tag(self, tagname: str) -> str:
        """
        Creates a tag with a given tagname

        Args:
            tagname: The name of the tag
        Returns:
            The id of the tag
        """
        if tagname not in self._tagids:
            if len(self._tags) == 0:
                new_id = "1"
            else:
                new_id = str(max(list(map(int, self._tags.keys()))) + 1)
            self._tags[new_id] = {"name": tagname, "files": []}
            self._tagids[tagname] = new_id
            return new_id
        else:
            return self._tagids[tagname]

    def file_in_db(self, filename: str) -> bool:
        """Check if a file is in the database"""
        return os.path.abspath(filename) in self._hashes

    def _hash_for_file(self, filename: str, store=False) -> str:
        """
        Get the hash for a file

        Args:
            filename: The filename
            store: If true the hash will be cached
        Returns:
            The hash
        """
        if filename in self._hashes:
            hash = self._hashes[filename]
        else:
            hash = self.compute_hash(filename)
            if store is True or (store == "is_file" and hash in self._files):
                self._hashes[filename] = hash
        return hash

    def get_or_create_file(self, filename: str, store=True) -> typing.Tuple[str, dict]:
        """
        Retrieve or create a file from the database

        Args:
            filename: The path to the file
            store: If True, a newly created object will be put into the database, else it isonly returned

        Returns:
            A tuple with the hash and the file dictionary
        """
        filename = os.path.abspath(filename)
        hash = self._hash_for_file(filename, store=store)
        if hash in self._files:
            return (hash, self._files[hash])
        fobj = {"filename": filename, "tags": [], "comments": ""}
        if store:
            return (hash, self._files.setdefault(hash, fobj))
        else:
            return (hash, fobj)

    def add_tags_to_file(self, fname: str, tags: typing.Union[typing.List[str], str]):
        """
        Add tags to a given file

        Args:
            fname: Path to the file
            tags: tagname or list of tagnames to add to the file
        """
        if not isinstance(tags, (tuple, list)):
            tags = [tags]
        hash, fobj = self.get_or_create_file(fname)
        tagids = self.tag_id_from_name(tags, create=True)

        for tag in tagids:
            if tag not in fobj["tags"]:
                fobj["tags"].append(tag)
                self._tags[tag]["files"].append(hash)

    def tag_id_from_name(
        self, tags: typing.Union[typing.List[str], str], create: bool = False
    ) -> typing.List[str]:
        """
        Get the id of a tag from its name.
        Can create a tag + id if it does not exist.

        Args:
            tags: tagname or list of tagnames
            create: True to create a tag and id
        Returns:
            List of tagids corresponding to the given tagnames
        """
        if not isinstance(tags, (list, tuple)):
            tags = [tags]
        tagids = []
        for tag in tags:
            if tag in self._tagids:
                tagids.append(self._tagids[tag])
            elif create:
                tagids.append(self.create_tag(tag))
        return tagids

    def delete_tag(self, tag: str):
        """
        Delete a tag

        Args:
            tag: The tagname to delete
        """
        id = self.tag_id_from_name([tag])[0]
        for f in self._tags[id]["files"]:
            del self._files[f]["tags"][self._files[f]["tags"].index(id)]
        del self._tagids[tag]
        del self._tags[id]

    def remove_tags_from_file(self, fname: str, tags: typing.Union[typing.List[str], str]):
        """
        Remove tags from a file

        Args:
            fname: Path to the file
            tags: tagname or list of tagnames
        """
        if not isinstance(tags, (list, tuple)):
            tags = [tags]
        if not self.file_in_db(fname):
            return
        hash, fobj = self.get_or_create_file(fname, store=False)
        tagids = self.tag_id_from_name(tags, create=False)
        fobj["tags"] = [t for t in fobj["tags"] if t not in tagids]
        for tid in tagids:
            self._tags[tid]["files"] = [i for i in self._tags[tid]["files"] if i != hash]

    def tags_for_file(
        self, fname: str, filecount: bool = False
    ) -> typing.List[typing.Union[str, typing.List[str | int]]]:
        """
        Get the tags associated with a file

        Args:
            fname: Path to the file
        Returns:
            List of tagnames
        """
        hash, fobj = self.get_or_create_file(fname, store=False)
        tags = fobj["tags"]
        if not filecount:
            return [self._tags[id]["name"] for id in tags]
        else:
            return [[self._tags[id]["name"], len(self._tags[id]["files"])] for id in tags]

    def tags_for_mult_files(
        self, fname: typing.List[str], filecount: bool = False
    ) -> typing.List[typing.Union[str, typing.List[str | int]]]:
        """
        Get the tags associated with a list of file

        Args:
            fname: Path to the file
        Returns:
            List of tagnames
        """
        ftags = {}
        for f in fname:
            ftags[f] = self.tags_for_file(f)
        tags = {}
        for f in ftags:
            for t in ftags[f]:
                tag = tags.get(t, 0)
                tag += 1
                tags[t] = tag
        if filecount:
            return [(k, v) for k, v in tags.items()]
        else:
            return list(tags.keys())

    def files_for_tag(self, tagname: str) -> typing.List[str]:
        """
        Get the files for a given tag

        Args:
            tagname: The name of the tag
        Returns:
            List of filepaths associated with this tag
        """
        if not tagname in self._tagids:
            return []
        else:
            hashes = self._tags[self._tagids[tagname]]["files"]
            return [self._files[h]["filename"] for h in hashes]

    def save(self):
        """Save the database to the file it was opened from"""
        with open(self.db_file, "w") as f:
            json.dump(self._db, f, indent=4)

    def ensure_database_integrity(self):
        newhash = {}
        for fn, hash in list(self._hashes.items()):
            if hash in self._files:
                newhash[os.path.abspath(fn)] = hash
        self._db["filehashes"] = newhash
        self._hashes = newhash
        for hash in self._files:
            self._files[hash]["filename"] = os.path.abspath(self._files[hash]["filename"])
            if hash not in self._hashes.values():
                self._hashes[self._files[hash]["filename"]] = hash

        for h, f in self._files.items():
            if "filename" not in f:
                fn = list(self._hashes.keys())[list(self._hashes.values()).index(h)]
                f["filename"] = fn
            if "tags" not in f:
                f["tags"] = []
            if "comments" not in f:
                f["comments"] = ""


class TagModel(LogEnabled, QtCore.QAbstractItemModel):
    def __init__(self, *args, tagger: Tagger):
        super().__init__(*args)
        self.tagger = tagger
        self.root_item = QtWidgets.QTreeWidgetItem()
        self.filename = ""
        self._tag_cache = None
        self._prev_tag_cache = None
        self._prev_tag_cache_filename = -1

    def _tags(self):
        if self._tag_cache is not None:
            return self._tag_cache
        s = time.perf_counter()
        if self.filename == "":
            tags = []
        elif self.filename is None:
            tags = self.tagger.available_tags(filecount=True)
        elif isinstance(self.filename, list):
            tags = self.tagger.tags_for_mult_files(self.filename, filecount=True)
        else:
            tags = self.tagger.tags_for_file(self.filename, filecount=True)
        self._tag_cache = tags
        self.log.debug(f"Getting Tags took {time.perf_counter() - s} s | {id(self)}")
        return tags

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        r = QtCore.Qt.ItemDataRole
        if role in [r.DisplayRole, r.EditRole]:
            if index.row() < 0:
                return ""
            tags = self._tags()
            return tags[index.row()][index.column()]

    def rowCount(self, index):
        if index.row() == -1:
            return len(self._tags())
        else:
            return 0

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if section == 0:
                return "Tagname"
            if section == 1:
                return "# Files"

    def columnCount(self, index) -> int:
        return 2

    def index(self, row, col, parent=QtCore.QModelIndex()):
        return self.createIndex(row, col, parent)

    def parent(self, index):
        return QtCore.QModelIndex()

    def change_file(self, filename):
        self.layoutAboutToBeChanged.emit()
        self.log.debug(f"Changing file to {filename}")
        if filename == self.filename and self._tag_cache is not None:
            self.log.debug("Filename did not actually change")
            return
        ptagcache = self._prev_tag_cache
        ptagcachename = self._prev_tag_cache_filename
        self._prev_tag_cache = self._tag_cache
        self._prev_tag_cache_filename = self.filename
        if filename == ptagcachename:
            self._tag_cache = ptagcache
        else:
            self._tag_cache = None
        # if self.filename is None and filename is not None:
        # self.log.debug("changed but using non tag cache")
        # self._tag_cache = self._non_tag_cache
        self.filename = filename
        self.layoutChanged.emit()

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        return QtCore.Qt.ItemIsEditable | QtCore.QAbstractItemModel.flags(self, index)

    def setData(self, index, value, role) -> bool:
        old_name = self.data(index, role)
        new_name = value
        if new_name == '':
            self.layoutAboutToBeChanged.emit()
            self.tagger.delete_tag(old_name)
            self.layoutChanged.emit()
        else:
            self.tagger.rename_tag(old_name, new_name)
        self.tagger.save()
        return True

    def hasIndex(self, row: int, column: int, parent) -> bool:
        if row >= self.rowCount(parent):
            return False
        return True


class TagList(LogEnabled, QtWidgets.QTreeView):
    tag_changed = QtCore.Signal(str)

    def __init__(self, parent, tagger):
        super().__init__(parent)
        self._model = TagModel(tagger=tagger)
        self.setModel(self._model)
        self.showColumn(1)
        self.setAlternatingRowColors(True)
        # self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setAllColumnsShowFocus(True)
        self.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)

    def currentChanged(self, current, previous) -> None:
        tag = self._model.data(current)
        self.tag_changed.emit(tag)
        return super().currentChanged(current, previous)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        item = self.indexAt(event.pos())
        selected = self.selectionModel().isSelected(item)
        self.log.debug(f"Clicked on Tag that is selected={selected}")
        r = super().mousePressEvent(event)
        if ((item.row() == -1 and item.column() == -1)) or selected:
            if not self.selectionModel().hasSelection():  # if nothing is selected
                self.log.debug("Nothing selected")
                return r
            self.selectionModel().setCurrentIndex(
                QtCore.QModelIndex(), QtCore.QItemSelectionModel.Rows
            )
            self.setCurrentIndex(QtCore.QModelIndex())
            self.tag_changed.emit("")
        return r

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        # return super().mouseReleaseEvent(event)
        pass


class Comment(QtWidgets.QTextEdit):
    def __init__(self, mdstate=True):
        super().__init__()
        self._text = ""
        self.show_markdown = mdstate
        # self.textChanged.connect(self._updatetext)

    def _updatetext(self):
        self._text = self.toPlainText()

    def setText(self, text):
        self._text = text
        if self.hasFocus() or not self.show_markdown:
            super().setText(text)
        else:
            self.setMarkdown(text)

    def setShowMarkdown(self, state):
        self.show_markdown = state

        if not self.hasFocus():
            self.focusInEvent(None)
            self.focusOutEvent(None)

    def focusInEvent(self, ev):
        self.setText(self._text)
        self.textChanged.connect(self._updatetext)
        if ev is not None:
            return super().focusInEvent(ev)

    def focusOutEvent(self, ev):
        self.textChanged.disconnect(self._updatetext)
        if self.show_markdown:
            self.setMarkdown(self._text)
        if ev is not None:
            return super().focusOutEvent(ev)


class Extractor(QtCore.QThread):
    done = QtCore.Signal(list)

    def __init__(self):
        super().__init__()
        self.results = []

    def run(self) -> None:
        files = [r.url for r in self.results]
        self.done.emit(files)


class TagWidget(LogEnabled, QtWidgets.QWidget):
    tag_filter_changed = QtCore.Signal(list)
    tag_changed = QtCore.Signal(list)

    def __init__(self, parent, tagger, comments_as_markdown=True):
        from .pdfsearch import EnterLineEdit

        super().__init__(parent)
        self.path = None
        self.files = None
        self.tagger = tagger
        self.list_cfile = TagList(self, tagger)
        self.list_cfile.hideColumn(1)
        self.list_all_file = TagList(self, tagger)
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.switch_global_tags = Switch("All Tags", "Filter Files", start_state=Switch.State.Right)
        # self.switch_global_tags.setChecked(True)
        self.switch_global_tags.stateChanged.connect(self.switch_all)
        self.layout.addWidget(self.switch_global_tags)
        self.layout.addWidget(self.list_all_file)
        self.layout.addWidget(QtWidgets.QLabel("Tags for selected file"))
        self.layout.addWidget(self.list_cfile)
        self.layout.addWidget(QtWidgets.QLabel("New Tag: (press Enter)"))
        self.tag_input = EnterLineEdit()
        self.tag_input.enterPressed.connect(self.new_tag)
        self.tag_input.setEnabled(False)
        self.layout.addWidget(self.tag_input)
        self.list_cfile.tag_changed.connect(self.update_filter)
        self.list_all_file.tag_changed.connect(self.update_files)
        # self.setFixedWidth(150)
        self.comment = Comment(comments_as_markdown)
        self.comment.setEnabled(False)
        self.comment.textChanged.connect(lambda: self.comment.setStyleSheet(""))
        self.layout.addWidget(self.comment)
        self.show_markdown = QtWidgets.QCheckBox("Show Markdown")
        self.show_markdown.toggled.connect(self.comment.setShowMarkdown)
        self.show_markdown.setChecked(comments_as_markdown)
        self.layout.addWidget(self.show_markdown)
        self.save = QtWidgets.QPushButton("Save Comment")
        self.layout.addWidget(self.save)
        self.save.clicked.connect(self.save_comment)
        # This shows all tags with None at the beginning. Default is now switch on the right (filter)
        # self.list_all_file.model().change_file(None)
        self.list_all_file.model().change_file([])
        self.list_cfile.setMaximumHeight(150)

        self.switch_global_tags.setToolTip(
            "If this switch is in the left position,"
            " selecting from the tag list below will "
            "show all files associated with this tag.\n\n"
            "If this switch is in the right position, "
            "selecting a tag will filter the search results "
            "for files with this tag"
        )
        self.extract_filenames_thread = Extractor()
        self.extract_filenames_thread.done.connect(self._update_list)

    def clear_selection(self):
        self.list_cfile.clearSelection()
        self.list_all_file.clearSelection()

    def save_comment(self):
        text = self.comment._text
        self.tagger.set_comment(self.path, text)
        self.tagger.save()
        self.comment.setStyleSheet("background-color: green;")

    def change_file(self, path):
        self.comment.setDisabled(path in ["", None])
        self.tag_input.setDisabled(path in ["", None])
        self.path = path
        self.list_cfile.model().change_file(path)
        self._sync_all_to_cfile_selection()
        if path not in ["", None]:
            self.comment.setText(self.tagger.get_comment(path))
        else:
            self.comment.setText("")

    def _sync_all_to_cfile_selection(self):
        self.list_cfile.clearSelection()
        mod = self.list_cfile.model()
        tag = self.list_all_file.model().data(self.list_all_file.currentIndex())
        for i in range(mod.rowCount(QtCore.QModelIndex())):
            if mod.data(mod.index(i, 0)) == tag:
                self.list_cfile.blockSignals(True)
                self.list_cfile.setCurrentIndex(mod.index(i, 0))
                self.list_cfile.model().layoutChanged.emit()
                self.list_cfile.blockSignals(False)
                break

    def _sync_cfile_to_all_selection(self):
        if not self.switch_global_tags.state == Switch.State.Left:
            self.list_all_file.clearSelection()
            mod = self.list_all_file.model()
            tag = self.list_cfile.model().data(self.list_cfile.currentIndex())
            for i in range(mod.rowCount(QtCore.QModelIndex())):
                if mod.data(mod.index(i, 0)) == tag:
                    self.list_all_file.blockSignals(True)
                    self.list_all_file.setCurrentIndex(mod.index(i, 0))
                    self.list_all_file.model().layoutChanged.emit()
                    self.list_all_file.blockSignals(False)
                    break

    def new_tag(self):
        if self.path is None:
            return
        text = self.tag_input.text()
        self.tag_input.setText("")
        self.list_cfile.model().layoutAboutToBeChanged.emit()
        self.tagger.add_tags_to_file(self.path, text)
        self.tagger.save()
        self.list_cfile.model().layoutChanged.emit()
        self.tag_input.setText("")

    def update_list_of_files(self, results):
        self.extract_filenames_thread.results = results
        self.extract_filenames_thread.start()

    def _update_list(self, files):
        self.files = files
        self.update_all_list(show_all=self.switch_global_tags.state == Switch.State.Left)

    def update_all_list(self, show_all=True):
        if show_all:
            self.list_all_file.model().change_file(None)
        else:
            self.list_all_file.model().change_file(self.files)

    def switch_all(self, state):
        if not state:
            self.list_all_file.model().change_file(self.files)
            self.tag_changed.emit(('', [""]))
            self.tag_filter_changed.emit([])
        self.update_all_list(show_all=state)
        self.list_all_file.selectionModel().clearSelection()

    def update_filter(self, tag):
        if tag == "":
            files = []
        else:
            files = self.tagger.files_for_tag(tag)
        self.tag_filter_changed.emit(files)
        self._sync_cfile_to_all_selection()

    def update_files(self, tag):
        self.log.debug(f"All Tag List updated {tag}")
        if tag == "":
            files = []
        else:
            files = self.tagger.files_for_tag(tag)
        self.log.debug("Changing Files")
        if self.switch_global_tags.state == Switch.State.Left:
            self.tag_changed.emit((tag, files))
        else:
            self._sync_all_to_cfile_selection()
            self.tag_filter_changed.emit(files)

    def switch_to_all_tags(self):
        self.switch_global_tags.setState(True)

    def search_tag(self, tag):
        self.log.debug(f"Searching tag {tag}")
        mod = self.list_all_file.model()
        di = QtCore.QModelIndex()
        for i in range(mod.rowCount(di)):
            if tag == mod.data(mod.index(i, 0)):
                self.list_all_file.setCurrentIndex(mod.index(i, 0))
                # self.list_all_file.tag_changed.emit(tag)
                # not sure why this is needed, but the new current item is not highlighted otherwise
                # (just after a resize/repaint from external)
                self.list_all_file.model().layoutChanged.emit()
                break
        self.tag_changed.emit((tag, self.tagger.files_for_tag(tag)))
